_ENV = AIContext

-- to get the desired research path spread, we don't use the values in 
-- empire.ai_state.research_path_desires directly. Instead, we determine the actual
-- research path desires by arranging all of the research paths from most 
-- desired to least desired. This ensures that there is a good distribution
-- of research path desires.
--
-- importantly, when we arrange the paths, the spacing is not uniform. 
-- it depends on two things:
--
-- 1) How much of a commitment the research path is. This is specified by the
-- research path itself. paths with a commitment_tech that is more expensive
-- will be spaced out further than techs with low commitment.
--
-- 2) How many prerequisites the tech shares in common with all of the more desired
-- techs in the ranking. If a tech shares a lot of prereqs in common with a highly
-- desired tech, it will be pulled up in the ranking.
local function get_desired_research_path_spread(empire, candidates)
  ensure_research_path_desires(empire)
  local research_path_desires in empire.ai_state
  local tech in empire

  -- unfortunately we will need to call research_path_spacing() on every combination
  -- of possible research paths. luckily most of the processing cost is involved in
  -- getting the set of commitment techs, which is independent of the combination
  -- and thus can be memorized.
  local mem_get_commitment_tech_set = strong_memoizeNN | function(path)

    local info = RESEARCH_PATHS[path]
    local t = info:resolve_commitment_tech(tech)

    --it is important that the result here is independent of what might already be researched
    local tech_sets = tech:candidate_tech_sets(t, true)
    local i = get_min_key(tech_sets,|) function(i, set)
      return tech:techset_cost(set)
    end

    local complete_set = tech_sets[i] or {}
    complete_set[t] = true

    return complete_set
  end

  local mem_research_cost = strong_memoizeNN | function(t)
    return empire:research_cost(t)
  end

  local function research_path_spacing(path1, path2)
    local commitment_techs = union(
      mem_get_commitment_tech_set(path1),
      mem_get_commitment_tech_set(path2)
    )

    local spacing = 0
    for t ; commitment_techs
      spacing += mem_research_cost(t)
    end
    return spacing
  end

  local research_path_list = {}
  for path ; candidates
    research_path_list[#research_path_list+1] = path
  end

  table.sort(research_path_list,|) function(a,b)
    return research_path_desires[a] > research_path_desires[b]
  end

  local distance_from_leader = {}
  local max_distance = 0

  for i = 1, #research_path_list do
    local path = research_path_list[i]

    local distance = 0
    for j = 1, i-1 do
      distance += research_path_spacing(research_path_list[i], research_path_list[j])
    end

    distance_from_leader[path] = distance
    max_distance = max(max_distance, distance)
  end

  --should see a smooth transition from 1 to 0...
  -- for i, path in ipairs(research_path_list) do
  --   print(path, 1 - distance_from_leader[path]/max_distance)
  -- end

  desired_research_spread = {}
  for path, distance ; distance_from_leader
    local normalized = distance/max_distance
    desired_research_spread[path] = 4^(1 - normalized)
  end

  normalize_table(desired_research_spread)

  return desired_research_spread
end


local function spent_research_amount(empire, t)
  local tech in empire

  local tech_sets = tech:candidate_tech_sets(t, true)
  local i = get_min_key(tech_sets,|) function(i, set)
    return tech:techset_cost(set)
  end

  local tech_set = tech_sets[i]
  if not tech_set
    return 0
  end

  local spent_beakers = 0
  for u ; tech_set
    if tech.known[u]
      spent_beakers += empire:research_cost(u)
    end
  end

  return spent_beakers
end


function choose_next_research(empire)
  --print("choose_next_research", empire.name)

  -- get all candidate techs for each research path
  local candidates = {}

  for path, path_info ; RESEARCH_PATHS
    candidates[path] = empire.tech:get_candidates_for_research_path(path_info)
  end

  -- within each research path, choose the least expensive tech (adjusted by m)
  for path, path_techs ; candidates
    candidates[path] = get_min_key(path_techs,|) function(t,m)
      local research_time = empire:estimate_years(t)
      if research_time
        return research_time/m
      end
    end
  end

  --print | v2s | candidates
  if not next(candidates)
    return
  end

  -- obtain the desired spread of research amongst the different path types
  local target_spread = get_desired_research_path_spread(empire, candidates)

  -- figure out how many beakers have already been put into each path
  local current_spread = {}
  for path, tech ; candidates
    current_spread[path] = spent_research_amount(empire, tech)
    --print(path, current[path])
  end

  -- figure out how many beakers will be in each path if a given candidate is researched
  local potential = {}
  for path, tech ; candidates
    potential[path] = setmetatable(copy | current_spread, sparse_metatable)
    potential[path][path] += empire:research_cost(tech)
    normalize_table(potential[path])
  end

  -- choose the one that brings us closest to our desired spread
  local best = get_min_key(potential,|) function(k,v)
    local err=0
    for path ; candidates
      local delta = abs(target_spread[path] - v[path])
      err += delta
    end
    --print(err, k, v2s | v)
    return err
  end

  --print(..best)
  if best
    print(empire.name, "next research", best, candidates[best])
    return candidates[best]
  end
end



if lua_load_complete
  for i, empire ; all_empires
    if not empire.is_minor and not empire.eliminated
      print(empire.name)
      local possible_paths = {}
      for path ; empire.ai_state.research_path_desires?
        if RESEARCH_PATHS[path]:possible(empire.tech)
          possible_paths[path] = true
        end
      end
      local research_desires = get_desired_research_path_spread(empire, possible_paths)
      for path, value in rpairs(research_desires,|) function(a,b) 
          return research_desires[a] > research_desires[b]
        end
        print(path, value)
      end
    end
  end
end