
ResearchPath_Methods = ResearchPath_Methods or {}

-- entries in a research path can either be a tech name string, or a one_of() list
-- given a Tech object, this resolves those down to a single tech
local function resolve_entry(tech, entry)
  if type(entry) == 'table'
    local i = get_min_key(entry,|) function(i, t)
      if tech.possible[t]
        return tech.empire:research_cost(t)
      end
    end
    return entry[i?]
  end

  return entry
end

function ResearchPath_Methods.resolve_core_tech(path, tech, i)
  return resolve_entry(tech, path.core_techs[i])
end

function ResearchPath_Methods.resolve_commitment_tech(path, tech)
  return resolve_entry(tech, path.commitment_tech)
end

-- return true if the research path is available
function ResearchPath_Methods.possible(path, tech)
  -- optional-tech-only paths are always available. optional techs are only ever researched if they are possible anyways.
  if not next(path.core_techs)
    return true
  end

  -- otherwise, the research path is available as long as the first core tech is possible
  local first_core = path:resolve_core_tech(tech, 1)
  if first_core and (tech.possible[first_core] or tech.known[first_core])
    --print("not possible", path.name, ..first_core)
    return true
  end

  return false
end

-- gets the next core tech in the given research path's progression
function Tech.next_core_tech(tech, path)
  for i, entry in ipairs(path.core_techs) do
    local next_tech = path:resolve_core_tech(tech, i)

    if not next_tech or not tech.possible[next_tech]
      break
    end

    if not tech.known[next_tech]
      return next_tech
    end
  end
end

-- iterates all available optional techs
function Tech.available_optional_techs(tech, path)
  return function(s,t)
    t = next(s,t)
    while t and not tech:researchable(t)
      t = next(s,t)
    end
    return t
  end, path.optional_techs
end

local function get_best_prereq_set(tech, t)
  local prereq in tech.info[t]
  local tech_sets = tech:candidate_tech_sets(prereq)

  local i = get_min_key(tech_sets,|) function(i, set)
    return tech:techset_cost(set)
  end

  return tech_sets[i]
end

function Tech.get_candidates_for_research_path(tech, path)
  if not path:possible(tech)
    return
  end

  local candidates = {}

  -- if there are no core techs then return any available optional techs
  if not next(path.core_techs)
    local candidates = {}
    for optional in tech:available_optional_techs(path) do
      --print("optional tech for research path", path.name, optional)
      candidates[optional] = 1
    end
    return candidates
  end

  local next_core = tech:next_core_tech(path)
  if next_core
    if tech:researchable(next_core)
      --if we can research this tech, it is the next candidate
      candidates[next_core] = 1
    else
      --otherwise, add all researchable prerequsites
      for prereq ; get_best_prereq_set(tech, next_core)
        if tech:researchable(prereq)
          --print("next tech for research path", path.name, prereq)
          candidates[prereq] = 0.75
        end
      end
    end
  end

  --check optional techs, but only if the first tech on this path has been researched
  if tech.known[first_core]
    for optional in tech:available_optional_techs(path) do
      --print("optional tech for research path", path.name, optional)
      candidates[optional] = 1
    end
  end

  return candidates
end