
_G.RESEARCH_PATHS = {}

local function research_path(name)
  return function(t)
    setmetatable(t, {__index = ResearchPath_Methods})

    t.name = name
    t.core_techs = t.core_techs or {}
    t.optional_techs = t.optional_techs and set(unpack(t.optional_techs)) or {}
    t.flags = t.flags or {}

    assert(t.commitment_tech, "commitment_tech not set for " .. name)

    RESEARCH_PATHS[name] = t
  end
end

function one_of(...)
  return { ... }
end

-- core_techs:
-- these are techs that the AI will try to research in order, if it chooses to 
-- go down that research path. if any prerequisites are unmet it will try to 
-- research those first.
--
-- optional_techs:
-- these are techs that the AI will try to research if the become available.
-- they are ignored if their prerequisites are unmet. order does not matter.
--
-- commitment_tech: NOT IMPLEMENTED YET
-- this is a tech that represents the amount of effort we need to put into
-- a research path to make it worthwhile. It affects the 'spacing' when
-- calculating the AI's desired research spread.
--
-- if commitment_tech is expensive, then if the research path is desired the AI will
-- put a more disproportionate amount of reserach into it. Conversely, if the
-- path is undesired, the AI will ignore that path to an even greater extent 
-- than normal if the commitment_tech is expensive.
--
-- also, if a path's commitment_tech shares prerequisites in common with the commitment_tech
-- of more desired research paths, it will get pulled higher up in the desired
-- research spread.
--
-- see get_desired_research_path_spread() in choose_research.lua
--
-- contingent_on:
-- if specified, the desirability of this research path will be limited by that of
-- the specified research path. it can still be significiantly less desirable, if
-- it rolls poorly, but it will never be significantly more desirable than the
-- path it is contingent on.

research_path 'missiles' {
  commitment_tech = 'Fusion Warheads',
  core_techs = {
    'Guidance Systems',
    'Fusion Warheads',
    'Retargeting Systems',
    'Antimatter Warheads',
  },
  optional_techs = {
    'Retargeting Systems', --in case anti-missiles are researched before fusion
  },
  flags = set('COMBAT_TECH'),
}

research_path 'heavy_missiles' {
  commitment_tech = 'Fusion Warheads',
  core_techs = {
    one_of('Magazine Bays', 'Missile Cruisers'),
    'Heavy Missiles',
    'Fusion Torpedoes',
    'Retargeting Systems',
    'Antimatter Torpedoes',
  },
  optional_techs = {
    'Miniaturization',
    'Plasma Torpedoes',
  },
  flags = set('COMBAT_TECH'),
}

research_path 'kinetics' {
  commitment_tech = 'Mass Drivers',
  core_techs = {
    'Mass Drivers',
  },
  optional_techs = {
    'Superconductors',
    'Nanocomputers',
  },
  flags = set('COMBAT_TECH'),
}

research_path 'heavy_kinetics' {
  contingent_on = 'kinetics',
  commitment_tech = 'Heavy Drivers',
  core_techs = {
    one_of('Heavy Hulls', 'Miniaturization'),--make sure to get a ship with heavy mounts first
    'Heavy Drivers',
    'Railguns',
    'Force Beams',
  },
  optional_techs = {
    'Superconductors',
    'Nanocomputers',
  },
  flags = set('COMBAT_TECH'),
}

research_path 'lasers' {
  commitment_tech = 'Rapid Fire Lasers',
  core_techs = {
    'Lasers',
    'Rapid Fire Lasers', --now a core tech. it doubles your firepower, so its pretty essential
    'Turbolasers',
    'Primary Beams',
  },
  optional_techs = {
    'X-Ray Lasers',
    'Nanocomputers',
    'Artificial Singularities'
  },
  flags = set('COMBAT_TECH'),
}

research_path 'heavy_lasers' {
  --ensures that the desire for this path is limited if the desire for the contingent path is low
  --so you can have a high laser desire with a low heavy_laser desire, but not the other way around
  contingent_on = 'lasers',
  commitment_tech = 'X-Ray Lasers',
  core_techs = {
    one_of('Heavy Hulls', 'Miniaturization'),--make sure to get a ship with heavy mounts first
    'Heavy Lasers',
    'Rapid Fire Lasers',
    'Advanced Turbolasers',
  },
  optional_techs = {
    'X-Ray Lasers', --should this be a core tech? and if so, where in the queue?
    'Nanocomputers',
    'Artificial Singularities'
  },
  flags = set('COMBAT_TECH'),
}

research_path 'ion_beams' {
  commitment_tech = 'Ion Beams', --if we choose ion beams as a focus, we really want to rush fusion and get ion beams
  core_techs = {
    'Ion Beams',
    'Disruptors',
    'Primary Beams',
  },
  optional_techs = {
    'Nanocomputers',
    'Artificial Singularities'
  },
  flags = set('COMBAT_TECH'),
}

research_path 'heavy_ions' {
  contingent_on = 'ion_beams',
  commitment_tech = 'Ion Beams',
  core_techs = {
    one_of('Heavy Hulls', 'Miniaturization'),--make sure to get a ship with heavy mounts first
    'Heavy Ion Beams',
    'Plasma Focusing',
    'Hellbores',
    'Artificial Singularities'
  },
  optional_techs = {
    'Nanocomputers',
    'Plasma Projectiles', --might as well grab it if we've researched heavy missiles
  },
  flags = set('COMBAT_TECH'),
}

--since the gremak viper dynamics tech doesn't really fit anywhere else
research_path 'viper_launchers' {
  commitment_tech = 'Viper Dynamics',
  core_techs = {
    'Viper Dynamics',
  },
}

research_path 'siege_ions' {
  --this may seem kind of strange. the idea is we want siege_ions to be available so that AI orthin
  --will research them, but if desired as a research path we don't want the AI to obsess over getting
  --to 'Stupendous Energy Physics' ASAP. So the effort tech is something along the way and useful
  -- but a little cheaper.
  commitment_tech = 'Ion Beams',
  core_techs = {
    'Stupendous Energy Physics',
    'Primary Artillery',
    'Stellar Surge Beam',
    'Artificial Singularities'
  },
  flags = set('COMBAT_TECH'),
}

research_path 'boarding_tactics' {
  commitment_tech = 'Docking Fields',
  core_techs = {
    --skipping the actual boarding tactics tech since the AI doesn't use it
    'Docking Fields', --just rush transports
    'Xeno-Engineering', --not sure if the AI actually uses this
    'Quantum Decryption',
  },
  optional_techs = {
    'Disruptors',
    'Neutron Beams',
    'EMP Warhead',
  },
  flags = set('COMBAT_TECH'),
}

research_path 'point_defenses' {
  commitment_tech = 'Close-In Weapons',
  optional_techs = {
    'Anti-Missiles',
    'Close-In Weapons',
    'Ion Point Defense',
    'Advanced Turbolasers',
    'Nanocomputers',
    'Escort Cruisers',
    'Escort Carrier',
    'Light Carrier',
    'Carriers',
  },
  flags = set('COMBAT_TECH'),
}

research_path 'armor' {
  commitment_tech = 'Duranium Construction',
  core_techs = {
    'Titanium Construction',
    'Duranium Construction',
    'Adamantium Construction',
    'Neutronium Construction',
  },
  flags = set('COMBAT_TECH'),
}

research_path 'shields' {
  commitment_tech = 'Deflector Screens',
  core_techs = {
    'Deflector Screens',
    'Battle Shields',
    'Force Fields',
    'Temporal Fields',
  },
  optional_techs = {
    'Shield Capacitors',
  },
  flags = set('COMBAT_TECH'),
}

research_path 'carriers' {
  --for ashdar, the carriers research path is much less of an all or nothing choice
  commitment_tech = one_of('Carriers', 'Light Carrier', 'Escort Carrier'),
  core_techs = {
    --allows ashdar to get their special carrier techs, in the right order
    one_of('Carriers', 'Light Carrier', 'Escort Carrier'),
    one_of('Carriers', 'Light Carrier'),
    'Carriers',
  },
  optional_techs = {
    'Multirole Fighters',
  },
  flags = set('COMBAT_TECH', 'SHIP_DESIGNS'),
}

research_path 'carrier_bombers' {
  contingent_on = 'carriers',
  commitment_tech = 'Fusion Warheads',
  core_techs = {
    one_of('Carriers', 'Light Carrier', 'Escort Carrier'), --need any carrier first
    'Bombers',
    'Fusion Warheads',
    'Antimatter Warheads',
  },
  optional_techs = {
    'Vindicators',
    'Multirole Fighters',
  },
  flags = set('COMBAT_TECH', 'SHIP_DESIGNS'),
}

research_path 'battleships' {
  commitment_tech = 'Heavy Hulls',
  core_techs = {
    'Heavy Hulls',
    one_of('Battlecruiser', 'Dreadnoughts'),
    'Dreadnoughts', --gremak have both Battlecruiser and Dreadnoughts techs
  },
  flags = set('COMBAT_TECH', 'SHIP_DESIGNS'),
}


research_path 'industry' {
  commitment_tech = 'Industrial Automation',
  core_techs = {
    'Industrial Automation',
    'Nanoscale Fabrication',
    'Subspace Networking',
    'Molecular Manipulation',
    'Industrial Replicators',
  },
  optional_techs = {
    'Adaptive Energy Grid',
    'Social Engineering',
    'Social Engineering',
    'Artificial Sentience',
    'Arda Fabrication Techniques',
  },
  flags = set('ECONOMY_TECH'),
}

research_path 'economy' {
  commitment_tech = 'Market Theory',
  core_techs = {
    'Market Theory',
    'Alien Psychology',
    'Interstellar Currency',
    'Galactic Stock Exchange'
  },
  optional_techs = {
    'Containment Fields',
    'Adaptive Energy Grid',
    'Social Engineering',
  },
  flags = set('ECONOMY_TECH'),
}

research_path 'science' {
  commitment_tech = 'Self-Organizing Networks',
  core_techs = {
    'Self-Organizing Networks',
    'Quantum Computing',
    'Hyperspace Cloud',
  },
  optional_techs = {
    --'Xenology', --not sure if the AI actually uses research modules
  },
  flags = set('ECONOMY_TECH'),
}

research_path 'expansion' {
  commitment_tech = 'Artificial Organisms',
  core_techs = {
    'Interstellar Expeditions',
    'Modular Outposts',
    'Artificial Organisms',
    'Hyperspace Beacons',
    'Bionomics',
    'Habitat Domes',
    'Weather Control',
    'Arcologies',
  },
  optional_techs = {
    'Alien Psychology',
    'Cloning',
    'Synthetic Foodstuffs',
    'Food Replicators',
    'Longevity',
    'Containment Fields',
    'Adaptive Energy Grid',
    'Social Engineering',
    'Graviton Physics',
    'Warp Lane Stabilizers',
  },
  flags = set('ECONOMY_TECH'),
}

research_path 'strategic_flexibility' {
  commitment_tech = 'Hyperspace Metrics',
  core_techs = {
    'Hyperspace Metrics',
    'Subspace Communications',
    'Hyperspace Beacons',
    'Graviton Physics',
    'Warp Lane Stabilizers',
    'Ansible',
  },
  optional_techs = {
    'Star Gates',
  },
}

research_path 'planetary_invasion' {
  commitment_tech = 'Hovertanks',
  core_techs = {
    'Hovertanks',
    'Battle Mechs',
    'Personal Shields',
  },
  optional_techs = {
    'Docking Fields', --this should become a core tech once regular transports get penalized for invasion
    'Advanced Field Theory',
  },
  flags = set('SECONDARY'),
}

research_path 'planetary_defense' {
  commitment_tech = 'Defense Grid',
  core_techs = {
    'Defense Grid',
    'Militia',
    'Hardened Bunkers',
    'Defense Network',
  },
  optional_techs = {
    'Advanced Field Theory',
    'Personal Shields',
    'Hovertanks',
    'Battle Mechs',
    'Orbital Construction',
    'Fleet Logistics',
    'Floating Fortresses',
  },
  flags = set('SECONDARY'),
}

-- if lua_load_complete
--   local empire = gui_player()
--   for path, info ; RESEARCH_PATHS
--     print(path, resolve_entry(empire.tech, info.commitment_tech), info:effort_beakers(empire.tech))
--   end
-- end

-- if lua_load_complete
--   for i, empire ; all_empires
--     if not empire.is_minor and not empire.eliminated
--       empire.ai_state.research_path_desires = nil
--     end
--   end
-- end